(function() {

    /**
     * Create/manage all Form embeds found on a page.
     */
    class EmbedManager {

        /**
         * Identify and process all DriveWorks Form embeds found in the page content.
         */
        processDOM() {
            const embedContainers = document.querySelectorAll("[data-driveworks-shortcode-embed]");
            if (!embedContainers) {
                return;
            }

            for (const container of embedContainers) {
                const embed = EmbedFactory.create(container);
                if (embed) {
                    embed.start();
                }
            }
        }
    }

    /**
     * Factory to create embed objects.
     */
    class EmbedFactory {
        static create(element) {
            const embed = new ShortcodeEmbed(element);

            // Get existing or create new client
            const client = clientManager.getOrCreateClient(element.dataset.serverUrl);
            if (!client) {
                embed.showError("Error creating DriveWorks Live Client. Check server_url is present and valid.", false);
                return;
            }
            embed.client = client;

            // Validate config values exist
            if (!embed.validateConfig()) {
                embed.showError("Invalid config.");
                return;
            }

            return embed;
        }
    }

    /**
     * Create/manage all DriveWorks Live clients.
     */
    class ClientManager {

        /**
         * Initializes an instance of the ClientManager.
         * @constructor
         */
        constructor() {
            this._clients = {};
        }

        /**
         * Retrieve or create DriveWorks Integration Theme client. Store for later use if created.
         * @param {string} serverUrl - URL of the Integration Theme server hosting the DriveWorks Live library.
         */
        getOrCreateClient(serverUrl) {
            if (!serverUrl) {
                return;
            }
            serverUrl = serverUrl.toLowerCase();

            // Check for pre-existing client using this url
            if (serverUrl in this._clients) {
                return this._clients[serverUrl];
            }

            try {
                // Create DriveWorks API client
                const client = new window.DriveWorksLiveClient(serverUrl); // use original?

                // Store client against url for re-use
                this._clients[serverUrl] = client;

                return client;
            } catch (error) {
                console.log(error);
            }
        }
    }

    /**
     * Individual DriveWorks Form embed.
     */
    class ShortcodeEmbed {

        /**
         * Initializes an instance of ShortcodeEmbed.
         * @constructor
         * @param {object} container - The HTML element containing connection details and form output.
         */
        constructor(container) {
            this._container = container;
            this._formData = container.dataset;
        }

        /**
         * Validate embed configuration.
         */
        validateConfig() {
            if (!this._client) {
                this.showError("No DriveWorks Live Client available.");
                return;
            }

            if (!this._formData.serverUrl) {
                this.showError("Please provide a valid server_url.", false);
                return false;
            }

            if (!this._formData.groupAlias) {
                this.showError("Please provide a valid group_alias.", false);
                return false;
            }

            return true;
        }

        /**
         * Set DriveWorks Live client for the embed to connect to the Integration Theme server.
         * @param {object} client - DriveWorks Live client object.
         */
        set client(client) {
            if (!client || this._client) {
                return;
            }

            this._client = client;
        }

        /**
         * Start individual DriveWorks Form embed. Login to Group and attach custom error handlers.
         */
        async start() {
            const groupErrorMsg = "Group login failed.";

            // Login to Group
            try {
                const groupAlias = this._formData.groupAlias;
                const session = await this._client.loginGroup(groupAlias);
                if (!session) {
                    throw groupErrorMsg;
                }
            } catch(error) {
                console.log(error);
                this.showError(groupErrorMsg);
                return;
            }

            // Add custom handling to client errors
            this._client.responseErrorDelegate = this.customClientErrorHandler;

            // DriveApps
            if (this._formData.driveAppAlias) {
                await this.runDriveApp();
                return;
            }

            // Projects
            await this.createSpecification();
        }

        /**
         * Create and render Specification Form. Setup timeout prevention method.
         */
        async createSpecification() {
            try {
                const groupAlias = this._formData.groupAlias;
                const projectName = this._formData.projectName;
                if (!projectName) {
                    this.showError("Please provide a valid project_name.", false);
                    return;
                }

                // Create new Specification
                const specification = await this._client.createSpecification(groupAlias, projectName);
                this.registerFormEvents(specification);

                // Find container, clear loading state
                const container = this._container.querySelector("[data-form-output]");
                container.innerHTML = "";

                // Render Specification
                await specification.render(container);
                this._container.classList.remove("dw-is-loading");

                // Prevent Specification timeout
                this.startPing(specification, this._formData.pingInterval);
            } catch (error) {
                console.log(error);
                this.showError("Error rendering Project.");
            }
        }

        /**
         * Run and render DriveApp Specification Form. Setup timeout prevention method.
         */
        async runDriveApp() {
            try {
                const groupAlias = this._formData.groupAlias;
                const driveAppAlias = this._formData.driveAppAlias;
                if (!driveAppAlias) {
                    this.showError("Please provide a valid drive_app_alias.", false);
                    return;
                }

                // Run DriveApp
                const driveApp = await this._client.runDriveApp(groupAlias, driveAppAlias);
                this.registerFormEvents(driveApp);

                // Find container, clear loading state
                const container = this._container.querySelector("[data-form-output]");
                container.innerHTML = "";

                // Render DriveApp
                await driveApp.render(container);
                this._container.classList.remove("dw-is-loading");

                // Prevent Specification timeout using regular ping
                this.startPing(driveApp, this._formData.pingInterval);
            } catch (error) {
                console.log(error);
                this.showError("Error rendering DriveApp.");
            }
        }

        /**
         * Register custom Form event listeners.
         * @param {object} specification - DriveWorks Specification object.
         */
        registerFormEvents(specification) {

            // Register a function to execute when the Specification is closed.
            specification.registerSpecificationClosedDelegate(() => {
                this.formClosed();
            });

            // Register a function to execute when the Specification is cancelled.
            specification.registerSpecificationCancelledDelegate(() => {
                this.formCancelled();
            });
        }

        /**
         * Send automated ping to Integration Theme server. Ensures Specification does not timeout from inactivity.
         * @param {object} specification - The Specification object.
         * @param {number} interval - The interval at which to ping, in seconds.
         */
        startPing(specification, interval) {
            if (!specification || !interval) { return; }

            // Disable ping if interval is not a number
            if (Number.isInteger(parseInt(interval)) === false) {
                console.error("ERROR: ping_interval is not a valid integer (in whole seconds)");
                return;
            }

            // Disable ping if interval is 0 or lower
            if (interval <= 0) {
                return;
            }

            // Start ping on interval
            this._pingInterval = setInterval(() => {
                specification.ping();
            }, interval * 1000);
        }

        /**
         * Form has closed (e.g. been completed), show message and clean up styling.
         */
        formClosed() {
            this._container.querySelector(".dw-message-form-complete").classList.add("message-shown");
            this.formCleanup();
        }

        /**
         * Form cancelled, show message and clean up styling.
         */
        formCancelled() {
            this._container.querySelector(".dw-message-form-cancel").classList.add("message-shown");
            this.formCleanup();
        }

        /**
         * Clean up client-side when form is cleared e.g. on cancel / close
         */
        formCleanup() {
            this.removeFormHeight();

            if (this._pingInterval) {
                clearInterval(this._pingInterval);
            }
        }

        /**
         * Remove any inline form height - revert container to 'natural' height based on its content.
         */
            removeFormHeight() {
            this._container.querySelector(".dw-form-output").style.height = "";
        }

        /**
         * Handle DriveWorks Live client errors with custom functions.
         * @param {object} response - Error response object.
         */
        customClientErrorHandler(response) {
            const statusCode = response.status;
            switch (statusCode) {

                // Unauthorized
                case 401:
                    this.showError("Authorization error. Confirm the default Group user can access this Project/DriveApp.", false);
                    return false;

                // Let the DriveWorks SDK handle all other errors
                default: return false;
            }
        }

        /**
         * Show Specification error messages on screen. Clear loading state.
         * @param {string} message - (optional) The custom error message to display.
         * @param {string} suffix - (optional) Append suffix to message.
         */
        showError(message, suffix = true) {
            const isAdmin = this._formData.debug;

            // Show generic error message when not logged in, or no message provided
            if (!isAdmin || !message) {
                message = "There has been a problem.";
                suffix = false;
            }

            this._container.innerHTML = `
                <div class="dw-embed-error">
                    ${isAdmin ? "<strong>[DriveWorks Shortcode Embed]</strong> " : ""}
                    ${message}
                    ${suffix ? " Please check the console." : ""}
                </div>
            `;
            this._container.classList.remove("dw-is-loading");
        }
    }

    // Create managers ready for use
    const embedManager = new EmbedManager();
    const clientManager = new ClientManager();

    // Detect external client library has loaded
    document.addEventListener("DOMContentLoaded", () => {
        const detectClientLoad = setInterval(() => {
            if (typeof window.DriveWorksLiveClient !== "undefined") {
                embedManager.processDOM();
                clearInterval(detectClientLoad);
                return;
            }
        }, 100);
    });
})();
